package com.taboola.flutter.advertisingId;

import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.content.pm.PackageManager;
import android.os.Handler;
import android.os.IBinder;
import android.os.IInterface;
import android.os.Looper;
import android.os.Parcel;
import android.os.RemoteException;
import android.text.TextUtils;
import android.util.Log;

import com.taboola.flutter.utils.SharedPrefUtil;

import java.io.IOException;
import java.util.concurrent.LinkedBlockingQueue;

public class AdvertisingIdInfo {
    private static final String TAG = AdvertisingIdInfo.class.getSimpleName();
    private String mAdvertisingId;
    private boolean mIsLimitedAdTrackingEnabled;

    public AdvertisingIdInfo(Context context) {
        init(context);
    }

    private void init(Context context) {
        Log.d(TAG, "init called");
        mAdvertisingId = getCachedAdvertisingId(context);
        updateAdvertisingIdAsync(context);
    }

    public String getAdvertisingId(Context context) {
        if (TextUtils.isEmpty(mAdvertisingId)) {
            Log.d(TAG, "getAdvertisingId: " + getCachedAdvertisingId(context));
            return getCachedAdvertisingId(context);
        } else {
            Log.d(TAG, "getAdvertisingId: " + mAdvertisingId);
            return mAdvertisingId;
        }
    }

    public boolean isLimitedAdTrackingEnabled() {
        return mIsLimitedAdTrackingEnabled;
    }

    private String getCachedAdvertisingId(Context context) {
        if (context == null) {
            Log.e(TAG, "getCachedAdvertisingId, context is null ", new Exception());
            return "";
        }
        return SharedPrefUtil.getCachedAdvertisingId(context);
    }

    public static final class AdInfo {
        private final String advertisingId;
        private final boolean limitAdTrackingEnabled;

        AdInfo(String advertisingId, boolean limitAdTrackingEnabled) {
            this.advertisingId = advertisingId;
            this.limitAdTrackingEnabled = limitAdTrackingEnabled;
        }

        public String getId() {
            return this.advertisingId;
        }
        boolean isLimitAdTrackingEnabled() {
            return this.limitAdTrackingEnabled;
        }
    }

    public void updateAdvertisingIdAsync(final Context context) {
        updateAdvertisingIdAsync(context, null);
    }


    /**
     * Attempts to update advertisingId on background thread. Callbacks are called on the UI Thread
     * <p>
     * {@link AdvertisingIdCallback#onIdRetrieved(String)} is called if advertisingId is
     * successfully retrieved
     * {@link AdvertisingIdCallback#onIdUnavailable()} is called if advertisingId is unavailable
     * (either because the "limit ad tracking" setting is {@code true} or an {@code Exception} was
     * thrown during the update) </p>
     */
    public void updateAdvertisingIdAsync(final Context context, final AdvertisingIdCallback callback) {
        Log.d(TAG, "updateAdvertisingIdAsync :: called");

        // need to be run on other thread to avoid exception from getAdvertisingIdInfo()
        new Thread(new Runnable() {
            @Override
            public void run() {
                Handler mainHandler = new Handler(context.getMainLooper());
                try {
                    final AdvertisingIdInfo.AdInfo adInfo = AdvertisingIdInfo.getAdvertisingIdInfo(context);
                    final String advertisingId = adInfo.getId();
                    Log.d(TAG, "updateAdvertisingIdAsync :: advertisingId = " + advertisingId);
                    mAdvertisingId = advertisingId;
                    mIsLimitedAdTrackingEnabled = adInfo.isLimitAdTrackingEnabled();
                    SharedPrefUtil.setCachedAdvertisingId(context, advertisingId);


                    mainHandler.post(new Runnable() {
                        @Override
                        public void run() {
                            if (callback != null) {
                                callback.onIdRetrieved(advertisingId);
                            }
                        }
                    });
                } catch (Exception e) {
                    Log.d(TAG, "updateAdvertisingIdAsync :: failed");
                    Log.e(TAG, "updateAdvertisingIdAsync :: failed", e);
                    mainHandler.post(new Runnable() {
                        @Override
                        public void run() {
                            if (callback != null) {
                                callback.onIdUnavailable();
                            }
                        }
                    });
                }
            }
        }).start();
    }

    private static AdvertisingIdInfo.AdInfo getAdvertisingIdInfo(Context context) throws Exception {
        if (Looper.myLooper() == Looper.getMainLooper())
            throw new IllegalStateException("Cannot be called from the main thread");

        try {
            PackageManager pm = context.getPackageManager();
            pm.getPackageInfo("com.android.vending", 0);
        } catch (Exception e) {
            Log.e(TAG, "Failed to get AdvertisingIdInfo", e);
            throw e;
        }

        AdvertisingIdInfo.AdvertisingConnection connection = new AdvertisingIdInfo.AdvertisingConnection();
        Intent intent = new Intent("com.google.android.gms.ads.identifier.service.START");
        intent.setPackage("com.google.android.gms");
        if (context.bindService(intent, connection, Context.BIND_AUTO_CREATE)) {
            try {
                AdvertisingIdInfo.AdvertisingInterface adInterface = new AdvertisingIdInfo.AdvertisingInterface(connection.getBinder());
                return new AdvertisingIdInfo.AdInfo(adInterface.getId(), adInterface.isLimitAdTrackingEnabled(true));
            } catch (Exception exception) {
                Log.e(TAG, "Failed ot get AdvertisingIdInfo", exception);
                throw exception;
            } finally {
                context.unbindService(connection);
            }
        } else {
            Log.e(TAG, "bindService fail");
            throw new IOException("Google Play connection failed");
        }
    }

    private static final class AdvertisingConnection implements ServiceConnection {
        boolean retrieved = false;
        private final LinkedBlockingQueue<IBinder> queue = new LinkedBlockingQueue<>(1);

        public void onServiceConnected(ComponentName name, IBinder service) {
            try {
                this.queue.put(service);
            } catch (InterruptedException localInterruptedException) {
                Log.e(TAG, "AdvertisingConnection | OnServiceConnected ", localInterruptedException);
            }
        }

        public void onServiceDisconnected(ComponentName name) {
        }

        IBinder getBinder() throws InterruptedException {
            if (this.retrieved) throw new IllegalStateException();
            this.retrieved = true;
            return this.queue.take();
        }
    }

    private static final class AdvertisingInterface implements IInterface {
        private IBinder binder;

        AdvertisingInterface(IBinder pBinder) {
            binder = pBinder;
        }

        public IBinder asBinder() {
            return binder;
        }

        public String getId() throws RemoteException {
            Parcel data = Parcel.obtain();
            Parcel reply = Parcel.obtain();
            String id;
            try {
                data.writeInterfaceToken("com.google.android.gms.ads.identifier.internal.IAdvertisingIdService");
                binder.transact(1, data, reply, 0);
                reply.readException();
                id = reply.readString();
            } finally {
                reply.recycle();
                data.recycle();
            }
            return id;
        }

        boolean isLimitAdTrackingEnabled(boolean paramBoolean) throws RemoteException {
            Parcel data = Parcel.obtain();
            Parcel reply = Parcel.obtain();
            boolean limitAdTracking;
            try {
                data.writeInterfaceToken("com.google.android.gms.ads.identifier.internal.IAdvertisingIdService");
                data.writeInt(paramBoolean ? 1 : 0);
                binder.transact(2, data, reply, 0);
                reply.readException();
                limitAdTracking = 0 != reply.readInt();
            } finally {
                reply.recycle();
                data.recycle();
            }
            return limitAdTracking;
        }
    }

    public interface AdvertisingIdCallback {
        void onIdRetrieved(String advertisingId);

        void onIdUnavailable();
    }
}